// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xbtree

import (
	"testing"
)

func TestRangeSet(t *testing.T) {
	const (
		oo  = KeyMax
		noo = KeyMin
	)

	type testEntry struct {
		A, B       *RangeSet
		Union      *RangeSet
		Difference *RangeSet
	}
	E := func(A, B, U, D *RangeSet) testEntry {
		return testEntry{A, B, U, D}
	}

	// S is shorthand to create RangeSet, e.g. S(1,2, 4,5) will return {[1,2) [4,5)}
	S := func(kv ...Key) *RangeSet {
		l := len(kv)
		if l % 2 != 0 {
			panic("odd number of keys")
		}
		S := &RangeSet{}
		for i := 0; i < l/2; i++ {
			// construct .rangev directly, not via AddRange
			lo := kv[2*i]
			hi := kv[2*i+1]
			hi_ := hi
			if hi_ != oo {
				hi_--
			}
			S.rangev = append(S.rangev, Range{lo, hi_})
		}
		S.verify()
		return S
	}

	testv := []testEntry{
		E(
			S(),	// A
			S(),	// B
				S(),	// U
				S()),	// D

		E(
			S(1,2),	// A
			S(1,2),	// B
				S(1,2),	// U
				S()),	// D

		// adjacent [1,3) [3,5)
		E(
			S(1,3), // A
			S(3,5), // B
				S(1,5),	 // U
				S(1,3)), // D

		// overlapping [1,3) [2,4)
		E(
			S(1,3),	// A
			S(2,4),	// B
				S(1,4),  // U
				S(1,2)), // D

		// [1,7) \ [3,5)  ->  [1,3) [5,7)
		E(
			S(1,7),	// A
			S(3,5),	// B
				S(1,7),
				S(1,3, 5,7)),

		// several ranges \ [-∞, ∞) -> ø
		E(
			S(1,3, 5,7, 11,100), // A
			S(noo, oo),          // B
				S(noo, oo),	// U
				S()),		// D

		// [1,3) [5,7) + insert [3,5) -> [1,7)
		E(
			S(1,3, 5,7),	// A
			S(3,5),		// B
				S(1,7),		// U
				S(1,3, 5,7)),	// D
	}

	for _, tt := range testv {
//		U := tt.A.Union(tt.B)
		D := tt.A.Difference(tt.B)

//		if !U.Equal(tt.Union) {
//			t.Errorf("Union:\n  A:   %s\n  B:   %s\n  ->u: %s\n  okU: %s\n", tt.A, tt.B, U, tt.Union)
//		}
		if !D.Equal(tt.Difference) {
			t.Errorf("Difference:\n  A:   %s\n  B:   %s\n  ->d: %s\n  okD: %s\n", tt.A, tt.B, D, tt.Difference)
		}
	}
}
